
/*
 * lcmaps-process-tracking
 * By Brian Bockelman, 2011 
 * This code is under the public domain
 */

/*****************************************************************************
                            Include header files
******************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/prctl.h>

#include "lcmaps/lcmaps_modules.h"
#include "lcmaps/lcmaps_cred_data.h"
#include "lcmaps/lcmaps_arguments.h"

#include "proc_police.h"
#include "proc_keeper.h"

// Various necessary strings
#define PLUGIN_PROCESS_TRACKING_PATH "-path"
char * logstr = "lcmaps-process-tracking";
char * execname = "/usr/share/lcmaps-plugins-process-tracking/process-tracking";

// Daemonize the process
int do_daemonize() {

    //  Setting the real and effective uid/gid to root.
    if (setreuid(0, 0) != 0) {
      lcmaps_log_debug(0, "%s: Setting the real/effective uid to root failed: %d %s.\n", logstr, errno, strerror(errno));
      return -errno;
    }
  
    if (setregid(0, 0) != 0) {
      lcmaps_log_debug(0, "%s: Setting the real/effective gid to root failed: %d %s.\n", logstr, errno, strerror(errno));
      return -errno;
    }

    int pid = fork();
    if (pid < 0) {
        lcmaps_log(0, "%s: Fork failure: %d %s", logstr, errno, strerror(errno));
        return -errno;
    }
    if (pid > 0) {
        _exit(0);
    }
    umask(0);
    int sid = setsid();
    if (sid < 0) {
        lcmaps_log(0, "%s: Setsid failure: %d %s", logstr, errno, strerror(errno));
        return -errno;
    }
    if ((chdir("/")) < 0) {
        lcmaps_log(0, "%s: Chdir failure: %d %s", logstr, errno, strerror(errno));
        return -errno; 
    }
    // Note that fd 0/1 are used to talk to the parent; keep them open.
    close(2);
    open("/dev/null", O_RDONLY);
    
    return 0;
}

#define PR_SET_NAME_MAX 16
int proc_police_main(pid_t pid, pid_t parent_pid) {
    int result = 0;

    lcmaps_log(0, "%s: Launching %s.\n", logstr, execname);

    char pid_char[64], ppid_char[64];
    if (snprintf(pid_char, 64, "%d", pid) >= 64) {
        lcmaps_log(0, "%s: Invalid PID: %d.\n", logstr, pid);
        return 1;
    }
    if (snprintf(ppid_char, 64, "%d", parent_pid) >= 64) {
        lcmaps_log(0, "%s: Invalid parent PID: %d.\n", logstr, parent_pid);
        return 1;
    }

    execl(execname, "process-tracking", pid_char, ppid_char, NULL);

    return result;
}

void handle_child(int p2c[], int c2p[], pid_t pid, pid_t ppid)
{
    // Close all file handles.
    //  Child Process
    close(p2c[1]);
    close(c2p[0]);
    if (dup2(p2c[0], 0) == -1) {
      lcmaps_log(0, "%s: Failed to dup file descriptor (%d: %s)\n", errno, strerror(errno));
      exit(errno);
    }
    if (dup2(c2p[1], 1) == -1) {
      lcmaps_log(0, "%s: Failed to dup file descriptor (%d: %s)\n", errno, strerror(errno));
      exit(errno);
    }
    close(p2c[0]);
    close(p2c[1]);
    close(c2p[1]);

    //  Setting the real and effective uid/gid to root.
    if (setreuid(0, 0) != 0) {
      lcmaps_log_debug(0, "%s: Setting the real/effective uid to root failed.\n", logstr);
      exit(errno);
    }

    if (setregid(0, 0) != 0) {
      lcmaps_log_debug(0, "%s: Setting the real/effective gid to root failed.\n", logstr);
      exit(errno);
    }
    if (do_daemonize()) {
      lcmaps_log_debug(0, "%s: Failed to daemonize!\n", logstr);
      exit(1);
    }
    
    int result = proc_police_main(pid, ppid);
    exit(result);
}

/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin; a no-op, but required by LCMAPS
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : success
******************************************************************************/
int plugin_initialize(int argc, char **argv)
{

  int idx;
  // Notice that we start at 1, as argv[0] is the plugin name.
  for (idx=1; idx<argc; argc++) {
    lcmaps_log_debug(2, "%s: arg %d is %s\n", logstr, idx, argv[idx]);
    if ((strncasecmp(argv[idx], PLUGIN_PROCESS_TRACKING_PATH, sizeof(PLUGIN_PROCESS_TRACKING_PATH)) == 0) && ((idx+1) < argc)) {
      if ((argv[idx+1] != NULL) && (strlen(argv[idx+1]) > 0)) {
        execname = strdup(argv[++idx]);
        lcmaps_log_debug(2, "%s: %s has %s\n", logstr, PLUGIN_PROCESS_TRACKING_PATH, execname);
      }
    } else {
      lcmaps_log(0, "%s: Invalid plugin option: %s\n", logstr, argv[idx]);
      return LCMAPS_MOD_FAIL;
    }
  }

  return LCMAPS_MOD_SUCCESS;

}


/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : success
******************************************************************************/
int plugin_introspect(int *argc, lcmaps_argument_t **argv)
{
  static lcmaps_argument_t argList[] = {
    {NULL        ,  NULL    , -1, NULL}
  };

  *argv = argList;
  *argc = lcmaps_cntArgs(argList);

  return LCMAPS_MOD_SUCCESS;
}




/******************************************************************************
Function:   plugin_run
Description:
    Launch a process tracking daemon for LCMAPS.
    Basic boilerplate for a LCMAPS plugin.
Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_run(int argc, lcmaps_argument_t *argv)
{
  FILE *fh = NULL;
  int p2c[2], c2p[2];
  int rc = 0, ok = 0;
  int uid_count;
  uid_t uid;
  pid_t pid, my_pid, ppid;

  uid_count = 0;
  uid_t * uid_array;
  uid_array = (uid_t *)getCredentialData(UID, &uid_count);
  if (uid_count != 1) {
    lcmaps_log(0, "%s: No UID set yet; must map to a UID before running the process tracking module.\n", logstr);
    goto process_tracking_uid_failure;
  }
  uid = uid_array[0];

  if (pipe(p2c) == -1) {
    lcmaps_log(0, "%s: Pipe creation failure (%d: %s)\n", errno, strerror(errno));
    goto process_tracking_pipe_failure;
  }
  if (pipe(c2p) == -1) {
    lcmaps_log(0, "%s: Pipe creation failure (%d: %s)\n", errno, strerror(errno));
    goto process_tracking_pipe_failure;
  }

  my_pid = getpid();
  ppid   = getppid();

  pid = fork();
  if (pid == -1) {
    lcmaps_log(0, "%s: Fork failure (%d: %s)\n", errno, strerror(errno));
    goto process_tracking_fork_failure;
  } else if (pid == 0) {
    handle_child(p2c, c2p, my_pid, ppid);
  }
  close(p2c[0]);
  close(p2c[1]);
  close(c2p[1]);

  //  After dumping the string to stdout, we need to expect an answer
  //  back on stdin.
  fh = fdopen(c2p[0], "r");
  if (fh == NULL) {
    rc = 1;
  } else {
    rc = fscanf(fh, "%d", &ok);
    fclose(fh);
  }

  // Child has exited, let's call waitpid and do status parsing.
  int status, exit_code, result;
  while (((result = waitpid(pid, &status, 0)) == -1) && (errno == EINTR)) {
  }
  if (result == -1) {
    lcmaps_log (0, "%s: Error: unable to get status of glexec monitor child. %d %s.\n", logstr, errno, strerror(errno));
    goto process_tracking_child_failure;
  }
  if (!WIFEXITED(status)) {
    lcmaps_log (0, "%s: Error: glexec monitor child died with a non-exit status, %d.\n", logstr, status);
    goto process_tracking_child_failure;
  }
  if ((exit_code = WEXITSTATUS(status))) {
    lcmaps_log (0, "%s: Error: glexec monitor child with exit code %d.\n", logstr, exit_code);
    goto process_tracking_child_failure;
  }

  if (rc < 0) {
    lcmaps_log (0, "%s: Error: failure reading from the monitor process. %d.\n", logstr, rc);
    goto process_tracking_child_failure;
  }
  if (ok != 0) {
    lcmaps_log (0, "%s: Error: failure in configuring monitor process. %d.\n", logstr, ok);
    goto process_tracking_child_failure;
  }

  lcmaps_log(0, "%s: monitor process successfully launched\n", logstr);

  return LCMAPS_MOD_SUCCESS;


process_tracking_fork_failure:
  close(p2c[0]);
  close(p2c[1]);
  close(c2p[0]);
  close(c2p[1]);
process_tracking_pipe_failure:
process_tracking_uid_failure:
process_tracking_child_failure:
  lcmaps_log_time(0, "%s: monitor process launch failed\n", logstr);

  return LCMAPS_MOD_FAIL;
}

int plugin_verify(int argc, lcmaps_argument_t * argv)
{
    return plugin_run(argc, argv);
}

/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin.  Boilerplate - doesn't do anything
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : success
******************************************************************************/
int plugin_terminate()
{
  return LCMAPS_MOD_SUCCESS;
}
